import os.path
import re
from collections.abc import Iterable
from contextlib import suppress
from textwrap import dedent
from typing import Any, ClassVar, Protocol, TextIO

from ast_grep_py import SgNode, SgRoot
from pydantic import BaseModel

from upcast.django_model import models
from upcast.utils import AnalysisModuleImport, FunctionArgs


class Plugin(Protocol):
    def should_run(self, context: models.Context, node: SgNode) -> bool: ...

    def run(self, context: models.Context, node: SgNode): ...

    def finish(self, context: models.Context) -> bool: ...


class Runner(BaseModel):
    context: models.Context

    class Config:
        arbitrary_types_allowed = True

    @property
    def plugins(self) -> list[Plugin]:
        return [
            ModuleImportPlugin(),
            ModelDefinitionPlugin(),
            ClassPathFixPlugin(),
            ExportModelPlugin(),
        ]

    def run(self, files: Iterable[TextIO]):
        plugins = self.plugins

        for f in files:
            self.context.switch_to_file(f.name)
            code = f.read()
            root = SgRoot(code, "python")
            root_node = root.root()

            for plugin in plugins:
                if plugin.should_run(self.context, root_node):
                    plugin.run(self.context, root_node)

        self.context.reset()
        while plugins:
            for plugin in plugins:
                if plugin.finish(self.context):
                    plugins.remove(plugin)


class FileBasePlugin(Plugin):
    def finish(self, context: models.Context) -> bool:
        return True


class ModuleImportPlugin(FileBasePlugin):
    """查找模块导入"""

    def should_run(self, context: models.Context, node: SgNode) -> bool:
        return "models" in context.current_file

    def run(self, context: models.Context, node: SgNode):
        analysis = AnalysisModuleImport(node=node, module=context.module)
        for module, name, alias in analysis.iter_import(node):
            context.add_imported_module(
                models.ImportedModule(
                    module=module,
                    name=name,
                    alias=alias,
                ),
                export=context.current_file.endswith(os.path.join("models", "__init__.py")),
            )


class ModelDefinitionPlugin(FileBasePlugin):
    """查找模型定义"""

    field_name_regex: ClassVar[re.Pattern] = re.compile(r"^\w[\w_]+$")
    field_type_regex: ClassVar[re.Pattern] = re.compile(r"^.*Field.*$")
    field_available_kwargs: ClassVar[set[str]] = {
        "primary_key",
        "db_index",
        "unique",
        "null",
        "blank",
        "choices",
        "help_text",
        "verbose_name",
        "max_length",
        "auto_created",
        "on_delete",
        "related_name",
    }

    def should_run(self, context: models.Context, node: SgNode) -> bool:
        return "models" in context.current_file

    def iter_base_classes(self, context: models.Context, node: SgNode, model: models.Model):
        for i in node.get_multiple_matches("BASE"):
            kind = i.kind()
            if kind in ["attribute", "identifier"]:
                yield models.ModelBase(node=i, name=i.text(), class_path=i.text())

    def is_look_like_django_field(self, attr: str, cls: str, kwargs: dict[str, Any]) -> bool:
        if not self.field_name_regex.match(attr):
            return False

        if not self.field_type_regex.match(cls):
            return False

        if kwargs.keys() & self.field_available_kwargs:
            return True

        return not kwargs

    def iter_fields(self, context: models.Context, node: SgNode, model: models.Model):
        class_indent = node.range().start.column

        for i in node.find_all(pattern="$ATTR = $CLS($$$ARGS)"):
            if i.range().start.column <= class_indent:
                continue

            attr = i["ATTR"].text()
            cls = i["CLS"].text()
            args = FunctionArgs().parse_args(i, "ARGS")

            if not self.is_look_like_django_field(attr, cls, args):
                continue

            imported, _, type_ = cls.partition(".")
            safe_kwargs = {}
            for k in self.field_available_kwargs:
                v = args.get(k)
                if not v:
                    continue

                with suppress(Exception):
                    safe_kwargs[k] = v.value()

            yield models.ModelField(
                node=i,
                name=attr,
                type=type_ or imported,
                class_path=cls,
                kwargs=safe_kwargs,
            )

    def iter_indexes_from_model(self, context: models.Context, node: SgNode, model: models.Model):
        for f in model.fields:
            kwargs = f.kwargs
            if kwargs.get("primary_key") == "True" or f.name == "id":
                yield models.ModelIndex(
                    node=f.node,
                    fields=[f.name],
                    kind="primary_key",
                )

            if kwargs.get("db_index") == "True":
                yield models.ModelIndex(
                    node=f.node,
                    fields=[f.name],
                    kind="index",
                )

            if kwargs.get("unique") == "True":
                yield models.ModelIndex(
                    node=f.node,
                    fields=[f.name],
                    kind="unique",
                )

    def iter_indexes_from_meta(self, context: models.Context, node: SgNode, model: models.Model):
        meta_cls = node.find(
            pattern=dedent(
                """
                class Meta:
                    $$$DEFINITION
                """
            )
        )
        if not meta_cls:
            return

        unique_together = meta_cls.find(pattern="unique_together = ($$$ARGS)")
        if unique_together:
            for i in unique_together.get_multiple_matches("ARGS"):
                yield models.ModelIndex(
                    node=i,
                    fields=[j.text() for j in i.get_multiple_matches("$$$ARGS") if j.kind() == "string"],
                    kind="unique",
                )

        indexes = meta_cls.find(pattern="indexes = [$$$INDEX]")
        if indexes:
            for i in indexes.get_multiple_matches("INDEX"):
                item = i.find(pattern="$F($$$ARGS)")
                if not item:
                    continue

                args = FunctionArgs().parse_args(item, "ARGS")
                fields = args["fields"].value()
                if not fields:
                    continue

                yield models.ModelIndex(
                    node=i,
                    fields=args["fields"].value(),
                    kind="unique" if args.get("unique") == "True" else "index",
                )

    def iter_indexes(self, context: models.Context, node: SgNode, model: models.Model):
        yield from self.iter_indexes_from_model(context, node, model)
        yield from self.iter_indexes_from_meta(context, node, model)

    def is_look_like_django_model(self, model: models.Model) -> bool:
        if not model.fields:
            return False

        if not model.bases:
            return False

        return any("Model" in m.name for m in model.bases)

    def locations(self, context: models.Context, node: SgNode, model: models.Model):
        yield f"{context.module}.{model.name}"

    def run(self, context: models.Context, node: SgNode):
        for i in node.find_all(
            pattern=dedent(
                """
                class $NAME($$$BASE):
                    $$$DEFINITION
                """
            )
        ):
            node_range = i.range()

            model = models.Model(
                node=i,
                name=i["NAME"].text(),
                file=context.current_file,
                position=(node_range.start.line, node_range.start.column),
            )

            model.bases.extend(self.iter_base_classes(context, i, model))
            model.fields.extend(self.iter_fields(context, i, model))
            model.indexes.extend(self.iter_indexes(context, i, model))
            model.locations.extend(self.locations(context, i, model))

            if not self.is_look_like_django_model(model):
                continue

            context.unresolved_models.append(model)

    def finish(self, context: models.Context) -> bool:
        for i in context.unresolved_models:
            for location in i.locations:
                context.resolved_models[location] = i

        context.unresolved_models = []
        return True


class ClassPathFixPlugin(FileBasePlugin):
    """修复类路径"""

    def should_run(self, context: models.Context, node: SgNode) -> bool:
        return True

    def get_module_path(self, context: models.Context, class_path: str):
        imported, _, attribute = class_path.partition(".")
        module = context.module_imports.get(imported)
        if module:
            return f"{module.module}.{class_path}"

        return f"{context.module}.{class_path}"

    def fix_bases(self, context: models.Context, model: models.Model):
        for i in model.bases:
            i.class_path = self.get_module_path(context, i.class_path)

    def fix_fields(self, context: models.Context, model: models.Model):
        for i in model.fields:
            i.class_path = self.get_module_path(context, i.class_path)

    def run(self, context: models.Context, node: SgNode):
        for model in context.unresolved_models:
            if model.file != context.current_file:
                continue

            self.fix_bases(context, model)
            self.fix_fields(context, model)


class ExportModelPlugin(FileBasePlugin):
    def should_run(self, context: models.Context, node: SgNode) -> bool:
        return False

    def finish(self, context: models.Context) -> bool:
        for model in context.resolved_models.values():
            for location in model.locations:
                module, _, name = location.rpartition(".")
                module_refs = context.module_refs.get(module)
                if not module_refs:
                    continue

                for imported in [name, "*"]:
                    module_ref = module_refs.get(imported)
                    if not module_ref:
                        continue

                    model.locations.append(f"{module_ref.module}.{module_ref.alias or name}")

        return True



import json
import sys
from contextlib import nullcontext
from glob import glob
from typing import Optional

import click

from upcast.django_model.core import Runner
from upcast.django_model.models import Context
from upcast.env_var.exporter import BaseExporter, MultiExporter
from upcast.env_var.plugin import EnvVarHub


@click.group()
def main():
    pass


@main.command()
@click.option("-o", "--output", default=[], type=click.Path(), multiple=True)
@click.option("--additional-patterns", default=[], multiple=True)
@click.option("--additional-required-patterns", default=[], multiple=True)
@click.argument("path", nargs=-1)
def find_env_vars(
    output: list[str],
    path: list[str],
    additional_patterns: list[str],
    additional_required_patterns: list[str],
):
    def iter_files():
        for i in path:
            with open(i) as f:
                yield f

    exporter = MultiExporter.from_paths(output) if output else BaseExporter()

    hub = EnvVarHub(
        exporter=exporter,
        additional_patterns=additional_patterns,
        additional_required_patterns=additional_required_patterns,
    )
    hub.run(iter_files())


@main.command()
@click.option("-o", "--output", default=None, type=click.Path())
@click.argument("root")
def analyze_django_models(output: Optional[str], root: str):
    def iter_files():
        for i in glob(f"{root}/**/*.py", recursive=True):
            with open(i) as f:
                yield f

    output_file = open(output, "w") if output else nullcontext(sys.stdout)

    with output_file as f:
        runner = Runner(context=Context(root_dir=root))
        runner.run(iter_files())

        f.write(
            json.dumps(
                [m.model_dump() for m in runner.context.get_models()],
                indent=2,
            )
        )


if __name__ == "__main__":
    main()



import csv
import os
from dataclasses import dataclass, field
from typing import TextIO

from pydantic import BaseModel

from upcast.env_var.core import EnvVar, EnvVarExporter


class BaseExporter(BaseModel):
    collected_vars: list[EnvVar] = field(default_factory=list)

    def begin(self):
        pass

    def handle(self, env_var: EnvVar):
        self.collected_vars.append(env_var)

    def get_merged_vars(self) -> dict[str, EnvVar]:
        merged_vars: dict[str, EnvVar] = {}
        for i in self.collected_vars:
            if i.name in merged_vars:
                merged_vars[i.name].merge_from(i, False)
            else:
                merged_vars[i.name] = i

        return merged_vars

    def end(self):
        pass


@dataclass
class MultiExporter:
    exporters: list[EnvVarExporter] = field(default_factory=list)

    @classmethod
    def from_paths(cls, paths: list[str]):
        exporters = []
        for i in paths:
            _, ext = os.path.splitext(i)
            if ext == ".csv":
                exporters.append(CSVExporter(path=i))
            elif ext == ".html":
                exporters.append(HTMLExporter(path=i))
            else:
                raise ValueError(f"Output format not supported: {ext}")

        return cls(exporters=exporters)

    def begin(self):
        for i in self.exporters:
            i.begin()

    def handle(self, env_var: EnvVar):
        for i in self.exporters:
            i.handle(env_var)

    def end(self):
        for i in self.exporters:
            i.end()


@dataclass
class CSVExporter:
    path: str
    file: TextIO = field(init=False)
    writer: csv.DictWriter = field(init=False)

    def __post_init__(self):
        self.file = open(self.path, "w", encoding="utf-8-sig")
        self.writer = csv.DictWriter(
            self.file,
            fieldnames=["name", "cast", "value", "required", "statement", "location"],
        )

    def begin(self):
        self.writer.writeheader()

    def handle(self, env_var: EnvVar):
        self.writer.writerow(
            {
                "name": env_var.name,
                "cast": env_var.cast,
                "value": env_var.value,
                "required": "*" if env_var.required else "",
                "location": env_var.location(),
                "statement": env_var.statement(),
            }
        )

    def end(self):
        self.file.close()


class ConsoleExporter(BaseExporter):
    def handle(self, env_var: EnvVar):
        prefix = ""
        if env_var.required:
            prefix = "*"

        if env_var.value:
            print(f"{prefix}{env_var.name}={env_var.value} at {env_var.location()}")
        else:
            print(f"{prefix}{env_var.name} at {env_var.location()}")


@dataclass
class HTMLExporter:
    path: str
    file: TextIO = field(init=False)

    def __post_init__(self):
        self.file = open(self.path, "w", encoding="utf-8")

    def begin(self):
        self.file.write("<html><head><title>Env Vars</title></head><body><table>")
        self.file.write(
            "\n".join(
                [
                    "<tr><th>Name</th>",
                    "<th>Value</th>",
                    "<th>Cast</th>",
                    "<th>Required</th>",
                    "<th>Statement</th></tr>" "<th>Location</th>",
                ]
            )
        )

    def handle(self, env_var: EnvVar):
        self.file.write(
            "\n".join(
                [
                    "<tr>",
                    f"<td>{env_var.name}</td>",
                    f"<td>{env_var.value}</td>",
                    f"<td>{env_var.cast}</td>",
                    f"<td>{'Yes' if env_var.required else ''}</td>",
                    f"<td>{env_var.statement()}</td>",
                    f"<td>{env_var.location()}</td>",
                    "</tr>",
                ]
            )
        )

    def end(self):
        self.file.write("</table></body></html>")
        self.file.close()



import os.path
from collections import defaultdict
from typing import Any

from ast_grep_py import SgNode
from pydantic import BaseModel, Field


class ModelField(BaseModel):
    node: SgNode = Field(title="Node of the field", exclude=True)
    name: str = Field("", title="Name of the field")
    type: str = Field("", title="Type of the field")
    class_path: str = Field("", title="Class path of the field")
    kwargs: dict[str, Any] = Field(default_factory=dict, title="Keyword arguments")

    class Config:
        arbitrary_types_allowed = True


class ModelIndex(BaseModel):
    node: SgNode = Field(title="Node of the index", exclude=True)
    fields: list[str] = Field(default_factory=list, title="Fields of the index")
    kind: str = Field("", title="Kind of the index")

    class Config:
        arbitrary_types_allowed = True


class ModelBase(BaseModel):
    node: SgNode = Field(title="Node of the model", exclude=True)
    name: str = Field("", title="Name of the model")
    class_path: str = Field("", title="Class path of the model")

    class Config:
        arbitrary_types_allowed = True


class Model(BaseModel):
    node: SgNode = Field(title="Node of the model", exclude=True)
    name: str = Field(title="Name of the model")
    file: str = Field(title="Path of the file")
    locations: list[str] = Field(default_factory=list, title="Modules of the model")
    position: tuple[int, int] = Field(title="Position of the model")
    bases: list[ModelBase] = Field(default_factory=list, title="Base classes of the model")
    fields: list[ModelField] = Field(default_factory=list, title="Fields of the model")
    indexes: list[ModelIndex] = Field(default_factory=list, title="Indexes of the model")

    class Config:
        arbitrary_types_allowed = True


class ImportedModule(BaseModel):
    module: str = Field("", title="Module name")
    name: str = Field("", title="Name of the class")
    alias: str = Field("", title="Alias of the class")


class Context(BaseModel):
    root_dir: str = Field(title="Root directory")
    module: str = Field("", title="Module name")
    current_file: str = Field("", title="Current file")
    resolved_models: dict[str, Model] = Field(default_factory=dict, title="Resolved models")
    unresolved_models: list[Model] = Field(default_factory=list, title="Unresolved models")
    module_imports: dict[str, ImportedModule] = Field(default_factory=dict, title="Imported models")
    module_refs: dict[str, dict[str, ImportedModule]] = Field(
        default_factory=lambda: defaultdict(dict), title="Module references"
    )

    class Config:
        arbitrary_types_allowed = True

    def switch_to_file(self, file: str):
        self.current_file = file if file else ""
        self.module = self.file_to_module(file)
        self.module_imports = {}

    def reset(self):
        self.switch_to_file("")

    def add_imported_module(self, module: ImportedModule, export: bool = False):
        self.module_imports[module.alias or module.name] = module

        if export:
            self.module_refs[module.module][module.name] = ImportedModule(
                module=self.module,
                name=module.name,
                alias=module.alias,
            )

    def file_to_module(self, file: str) -> str:
        if not file:
            return ""

        relpath = os.path.relpath(file, self.root_dir)
        path, file = os.path.split(relpath)

        if file == "__init__.py":
            pass
        elif "." not in file:
            path = relpath
        else:
            path, _ = os.path.splitext(relpath)

        return path.replace(os.path.sep, ".")

    def get_models(self) -> list[Model]:
        models: dict[int, Model] = {}

        for i in self.resolved_models.values():
            models[id(i)] = i

        for i in self.unresolved_models:
            models[id(i)] = i

        return list(models.values())



from dataclasses import dataclass, field
from typing import Any, Optional

import evalidate
from ast_grep_py import SgNode


@dataclass
class Evalidate:
    eval_model: evalidate.EvalModel = field(
        default_factory=lambda: evalidate.EvalModel(
            nodes=[
                "JoinedStr",
                "Expression",
                "FormattedValue",
                "Constant",
                "Name",
                "Load",
                "BinOp",
                "Add",
                "Sub",
                "Mult",
                "Div",
                "Mod",
                "Attribute",
                "List",
            ]
        )
    )

    def eval(self, code: str, scopes: Optional[dict[str, Any]] = None) -> Any:
        return evalidate.Expr(code, self.eval_model).eval(scopes)


@dataclass
class FunctionArg:
    name: str
    node: SgNode
    value_node: SgNode

    def value(self, **scopes) -> str:
        return Evalidate().eval(self.value_node.text(), scopes)


@dataclass
class FunctionArgs:
    args: dict[str, FunctionArg] = field(default_factory=dict)

    def parse(self, node: SgNode, group: str, args: tuple[str] = ()) -> "FunctionArgs":
        pos = 0
        len_args = len(args)

        for i in node.get_multiple_matches(group):
            kind = i.kind()

            if kind == ",":
                continue

            elif kind == "keyword_argument":
                pos = len_args
                name_node = i.child(0)
                name = name_node.text()

                self.args[name] = FunctionArg(name=name, node=i, value_node=i.child(2))

            elif pos < len_args:
                pos += 1
                name = args[pos]

                self.args[name] = FunctionArg(name=name, node=i, value_node=i)

        return self

    def parse_args(self, node: SgNode, group: str, args: tuple[str] = ()) -> dict[str, FunctionArg]:
        return self.parse(node, group, args).args


@dataclass
class AnalysisModuleImport:
    node: SgNode
    module: str

    def make_path_absolute(self, path: str) -> str:
        if not path.startswith("."):
            return path

        relpath = path.lstrip(".")
        module_list = self.module.rsplit(".", maxsplit=len(path) - len(relpath))
        return ".".join(module_list) + relpath

    def parse_multiple_matches(self, matches: list[SgNode]):
        for i in matches:
            node_kind = i.kind()
            if node_kind in ("dotted_name", "wildcard_import"):
                yield i.text(), ""

            elif node_kind == "aliased_import":
                # x as y
                yield i.child(0).text(), i.child(2).text()

    def iter_import_directly(self, node: SgNode):
        for i in node.find_all(pattern="import $$$NAME"):
            for path, alias in self.parse_multiple_matches(i.get_multiple_matches("NAME")):
                path, _, name = path.rpartition(".")
                yield path, name, alias

    def iter_import_from(self, node: SgNode):
        for i in node.find_all(pattern="from $MODULE import $$$NAME"):
            path = self.make_path_absolute(i["MODULE"].text())

            for name, alias in self.parse_multiple_matches(i.get_multiple_matches("NAME")):
                yield path, name, alias

    def iter_import(self, node: SgNode):
        yield from self.iter_import_directly(node)
        yield from self.iter_import_from(node)
